from rivalcfg import devices


class TestListPluggedDevice(object):
    def test_debug_devices(self, monkeypatch):
        monkeypatch.setenv("RIVALCFG_PROFILE", "1038:1702")
        devices_list = list(devices.list_plugged_devices())
        debug_device_found = False
        for device in devices_list:
            if (
                device["vendor_id"] == 0x1038
                and device["product_id"] == 0x1702
                and device["name"] == "SteelSeries Rival 100"
            ):
                debug_device_found = True
        assert debug_device_found



"""
This file contains the layout for multimedia keys.
"""


#: The multimedia keys layout
layout = {
    "Mute": 0xE2,
    "Next": 0xB5,
    "PlayPause": 0xCD,
    "Previous": 0xB6,
    "VolumeUp": 0xE9,
    "VolumeDown": 0xEA,
}

#: Alias for some keys of the layout
aliases = {
    "play": "PlayPause",
    "pause": "PlayPause",
    "prev": "Previous",
    "volup": "VolumeUp",
    "volume+": "VolumeUp",
    "vol+": "VolumeUp",
    "volumeplus": "VolumeUp",
    "volplus": "VolumeUp",
    "voldown": "VolumeDown",
    "voldwn": "VolumeDown",
    "voldn": "VolumeDown",
    "volume-": "VolumeDown",
    "vol-": "VolumeDown",
    "volumeminus": "VolumeDown",
    "volminus": "VolumeDown",
}



from .. import usbhid


profile = {
    "name": "SteelSeries Aerox 3",
    "models": [
        {
            "name": "SteelSeries Aerox 3",
            "vendor_id": 0x1038,
            "product_id": 0x1836,
            "endpoint": 3,
        },
    ],
    "settings": {
        "sensitivity": {
            "label": "Sensibility presets",
            "description": "Set sensitivity preset (DPI)",
            "cli": ["-s", "--sensitivity"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x2D],
            "value_type": "multidpi_range",
            "input_range": [200, 8500, 100],
            "output_range": [0x04, 0xC5, 2.33],
            "dpi_length_byte": 1,
            "first_preset": 1,
            "count_mode": "number",
            "max_preset_count": 5,
            "default": "800, 1600",
        },
        "polling_rate": {
            "label": "Polling rate",
            "description": "Set polling rate (Hz)",
            "cli": ["-p", "--polling-rate"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x2B],
            "value_type": "choice",
            "choices": {
                125: 0x04,
                250: 0x03,
                500: 0x02,
                1000: 0x01,
            },
            "default": 1000,
        },
        "z1_color": {
            "label": "Strip top LED color",
            "description": "Set the color of the top LED",
            "cli": ["--top-color", "--z1"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x21, 0x01],
            "value_type": "rgbcolor",
            "default": "red",
        },
        "z2_color": {
            "label": "Strip middle LED color",
            "description": "Set the color of the middle LED",
            "cli": ["--middle-color", "--z2"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x21, 0x02, 0x00, 0x00, 0x00],
            "value_type": "rgbcolor",
            "default": "lime",
        },
        "z3_color": {
            "label": "Strip bottom LED color",
            "description": "Set the color of the bottom LED",
            "cli": ["--bottom-color", "--z3"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x21, 0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00],
            "value_type": "rgbcolor",
            "default": "blue",
        },
        "reactive_color": {
            "label": "Reactive color",
            "description": "Set the color of the LEDs in reaction to a button click",
            "cli": ["-a", "--reactive-color"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x26],
            "value_type": "reactive_rgbcolor",
            "default": "off",
        },
        "led_brightness": {
            "label": "LED Brightness",
            "description": "Set the brightness of the LEDs",
            "cli": ["-l", "--led-brightness"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x23],
            "value_type": "range",
            "input_range": [0, 100, 1],
            "output_range": [0x00, 0x64, 1],
            "default": 100,
        },
        "rainbow_effect": {
            "label": "rainbow effect",
            "description": "Set the rainbow effect (can be cleared by setting a color)",
            "cli": ["-e", "--rainbow-effect"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x22],
            "value_type": "choice",
            # fmt: off
            "choices": {
                "all":           0b111,
                "bottom":        0b100,
                "middle":        0b010,
                "top":           0b001,
                "bottom-middle": 0b110,
                "middle-top":    0b011,
                "bottom-top":    0b101,
            },
            # fmt: on
            "default": "all",
        },
        "buttons_mapping": {
            "label": "Buttons mapping",
            "description": "Set the mapping of the buttons",
            "cli": ["-b", "--buttons"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x2A],
            "value_type": "buttons",
            # fmt: off
            "buttons": {
                "Button1":    {"id": 0x01, "offset": 0x00, "default": "button1"},
                "Button2":    {"id": 0x02, "offset": 0x05, "default": "button2"},
                "Button3":    {"id": 0x03, "offset": 0x0A, "default": "button3"},
                "Button4":    {"id": 0x04, "offset": 0x0F, "default": "button4"},
                "Button5":    {"id": 0x05, "offset": 0x14, "default": "button5"},
                "Button6":    {"id": 0x06, "offset": 0x19, "default": "dpi"},
                "ScrollUp"  : {"id": 0x31, "offset": 0x1E, "default": "scrollup"},
                "ScrollDown": {"id": 0x32, "offset": 0x23, "default": "scrolldown"},
            },
            "button_field_length": 5,
            "button_disable":     0x00,
            "button_keyboard":    0x51,
            "button_multimedia":  0x61,
            "button_dpi_switch":  0x30,
            "button_scroll_up":   None,
            "button_scroll_down": None,
            # fmt: on
            "default": "buttons(button1=button1; button2=button2; button3=button3; button4=button4; button5=button5; button6=dpi; scrollup=scrollup; scrolldown=scrolldown; layout=qwerty)",
        },
    },
    "save_command": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "command": [0x11, 0x00],
    },
}



"""
This module contains varous helper functions related to color.
"""


import re


# fmt: off
NAMED_COLORS = {
    "white":   (0xFF, 0xFF, 0xFF),
    "silver":  (0xC0, 0xC0, 0xC0),
    "gray":    (0x80, 0x80, 0x80),
    "black":   (0x00, 0x00, 0x00),
    "red":     (0xFF, 0x00, 0x00),
    "maroon":  (0x80, 0x00, 0x00),
    "yellow":  (0xFF, 0xFF, 0x00),
    "olive":   (0x80, 0x80, 0x00),
    "lime":    (0x00, 0xFF, 0x00),
    "green":   (0x00, 0x80, 0x00),
    "aqua":    (0x00, 0xFF, 0xFF),
    "teal":    (0x00, 0x80, 0x80),
    "blue":    (0x00, 0x00, 0xFF),
    "navy":    (0x00, 0x00, 0x80),
    "fuchsia": (0xFF, 0x00, 0xFF),
    "purple":  (0x80, 0x00, 0x80),
}
# fmt: on


def is_color(string):
    """Checks if the given string is a valid color.

    :param str string: The string to check.
    :rtype: bool

    >>> from rivalcfg.color_helpers import is_color
    >>> is_color("#FF0000") # hexadecimal colors are supported
    True
    >>> is_color("#F00")  # short format allowed
    True
    >>> is_color("FF0000")  # "#" is optionnal
    True
    >>> is_color("#ff0000")  # case insensitive
    True
    >>> is_color("#FF00")  # not a valid color
    False
    >>> is_color("red")  # named color are supported
    True
    >>> is_color("RED")  # named color are case insensitive
    True
    >>> is_color("hello")  # not a valid color name
    False
    """
    if string.lower() in NAMED_COLORS:
        return True
    if re.match(r"^#?[0-9a-f]{3}([0-9a-f]{3})?$", string, re.IGNORECASE):
        return True
    return False


def parse_color_string(color):
    """Converts a color string into an RGB tuple.

    :param str color: The string to convert.
    :return: (r, g, b)

    >>> from rivalcfg.color_helpers import parse_color_string
    >>> parse_color_string("#FF0000") # hexadecimal colors are supported
    (255, 0, 0)
    >>> parse_color_string("#F00")  # short format allowed
    (255, 0, 0)
    >>> parse_color_string("FF0000")  # "#" is optionnal
    (255, 0, 0)
    >>> parse_color_string("#ff0000")  # case insensitive
    (255, 0, 0)
    >>> parse_color_string("red")  # named color are supported
    (255, 0, 0)
    >>> parse_color_string("RED")  # named color are case insensitive
    (255, 0, 0)
    """
    # Named color
    if color.lower() in NAMED_COLORS:
        return NAMED_COLORS[color.lower()]

    # #f00 or #ff0000 -> f00 or ff0000
    if color.startswith("#"):
        color = color[1:]

    # f00 -> ff0000
    if len(color) == 3:
        color = color[0] * 2 + color[1] * 2 + color[2] * 2

    # ff0000 -> (255, 0, 0)
    return (
        int(color[0:2], 16),
        int(color[2:4], 16),
        int(color[4:], 16),
    )


def parse_color_gradient_string(gradient):
    """Parse a color gradient string.

    :param str gradient: The gradient string.
    :rtype: list

    >>> parse_color_gradient_string("0%: red, 33%: #00ff00, 66: 00f")
    [{'pos': 0, 'color': (255, 0, 0)}, {'pos': 33, 'color': (0, 255, 0)}, {'pos': 66, 'color': (0, 0, 255)}]
    >>> parse_color_gradient_string("-1%: red")
    Traceback (most recent call last):
        ...
    ValueError: invalid color stop position '-1%'
    >>> parse_color_gradient_string("150: red")
    Traceback (most recent call last):
        ...
    ValueError: invalid color stop position '150%'
    >>> parse_color_gradient_string("42%: hello")
    Traceback (most recent call last):
        ...
    ValueError: invalid color 'hello'
    >>> parse_color_gradient_string("hello")
    Traceback (most recent call last):
        ...
    ValueError: invalid color gradient 'hello'. ...
    """
    gradient = gradient.replace(" ", "").replace("%", "")

    if not re.match(r"[0-9-]+:[a-zA-Z0-9#]+(,[0-9]+:[a-zA-Z0-9#]+)*", gradient):
        raise ValueError(
            "invalid color gradient '%s'. It must looks like '<POS1>:<COLOR1>,<POS2>:<COLOR2>,...'"
            % gradient
        )

    result = []
    for pos, color in [s.split(":") for s in gradient.split(",")]:
        pos = int(pos)

        if not 0 <= pos <= 100:
            raise ValueError("invalid color stop position '%i%%'" % pos)

        if not is_color(color):
            raise ValueError("invalid color '%s'" % color)

        result.append(
            {
                "pos": pos,
                "color": parse_color_string(color),
            }
        )

    return result



import sys
import os.path
import platform

import hid
from pkg_resources import get_distribution
from .version import VERSION
from . import udev
from .mouse import get_mouse


def _make_title(text):
    separator = "=" * len(text)
    return "\n%s\n%s\n" % (text.upper(), separator)


def _get_os_info():
    result = _make_title("Operating System")
    if platform.system() == "Linux":
        result += _get_os_linux_info()
    else:
        result += "OS: %s" % platform.system()
    return result


def _get_os_linux_info():
    result = ""
    result += "OS: %s\n" % platform.system()
    result += "Platform: %s\n" % platform.platform()
    result += "Version: %s\n" % platform.version()
    if os.path.isfile("/etc/issue"):
        try:
            with open("/etc/issue", "r") as distro_file:
                distro = distro_file.read().strip()
                result += "Distribution issue: %s\n" % distro
        except Exception:
            pass
    return result


def _get_rivalcfg_info():
    result = _make_title("Rivalcfg")
    result += "Version: %s\n" % VERSION
    if platform.system() == "Linux":
        result += "udev rules installed: %a\n" % os.path.isfile(udev.RULES_FILE_PATH)
        result += "udev rules up to date: %a\n" % udev.is_rules_file_up_to_date()
    result += "Installation path: %s\n" % os.path.dirname(__file__)
    return result


def _get_python_info():
    result = _make_title("Python")
    result += "Python version: %d.%d.%d\n" % sys.version_info[:3]
    result += "HIDAPI version: %s\n" % get_distribution("hidapi").version
    return result


def _get_plugged_device_list():
    result = _make_title("Plugged SteelSeries devices endpoints")
    for device in hid.enumerate(0x1038):
        firmware_version = "0"
        try:
            with get_mouse(device["vendor_id"], device["product_id"]) as mouse:
                firmware_version = mouse.firmware_version
        except Exception:
            pass
        result += "%04x:%04x | %02x | %s (firmware v%s)\n" % (
            device["vendor_id"],
            device["product_id"],
            device["interface_number"],
            device["product_string"],
            firmware_version,
        )
    return result


def get_debug_info():
    result = ""
    result += _get_rivalcfg_info()
    result += _get_os_info()
    result += _get_python_info()
    result += _get_plugged_device_list()
    return result



from . import none
from . import choice
from . import range
from . import rgbcolor
from . import reactive_rgbcolor
from . import rgbgradient
from . import rgbgradientv2
from . import multidpi_range
from .buttons import buttons

# flake8: noqa



from .. import usbhid


profile = {
    "name": "SteelSeries Prime",
    "models": [
        {
            "name": "SteelSeries Prime",
            "vendor_id": 0x1038,
            "product_id": 0x182E,
            "endpoint": 0,
        },
        {
            "name": "SteelSeries Prime Rainbow 6 Siege Black Ice Edition",
            "vendor_id": 0x1038,
            "product_id": 0x182A,
            "endpoint": 0,
        },
        {
            "name": "SteelSeries Prime CS:GO Neo Noir Edition",
            "vendor_id": 0x1038,
            "product_id": 0x1856,
            "endpoint": 0,
        },
    ],
    "settings": {
        "sensitivity": {
            "label": "Sensitivity presets",
            "description": "Set sensitivity preset (DPI)",
            "cli": ["-s", "--sensitivity"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x61],
            "value_type": "multidpi_range",
            "input_range": [50, 18000, 50],
            "output_range": [1, 0x0168, 1],
            "dpi_length_byte": 2,
            "count_mode": "number",
            "first_preset": 0,
            "max_preset_count": 5,
            "default": "400, 800, 1200, 2400, 3200",
        },
        "polling_rate": {
            "label": "Polling rate",
            "description": "Set polling rate (Hz)",
            "cli": ["-p", "--polling-rate"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x5D],
            "value_type": "choice",
            "choices": {
                125: 0x04,
                250: 0x03,
                500: 0x02,
                1000: 0x01,
            },
            "default": 1000,
        },
        "color": {
            "label": "Wheel LED color",
            "description": "Set the color of the wheel LED",
            "cli": ["-c", "--color"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x62, 0x01],
            "command_suffix": [
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0x00,
                0xFF,
            ],
            "value_type": "rgbcolor",
            "default": "#FF5200",
        },
        "led_brightness": {
            "label": "Wheel LED brightness",
            "description": "Set the brightness of the wheel LED",
            "cli": ["-l", "--led-brightness"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x5F],
            "value_type": "range",
            "input_range": [0, 256, 1],
            "output_range": [0, 256, 1],
            "range_length_byte": 2,
            "default": 256,
        },
    },
    "save_command": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "command": [0x59],
    },
}



import os
import json


def get_xdg_config_home():
    """Returns the path of the folder where to store the configs (generally
    ``$HOME/.config``).

    :rtype: str
    """
    if "XDG_CONFIG_HOME" in os.environ and os.environ["XDG_CONFIG_HOME"]:
        return os.environ["XDG_CONFIG_HOME"]
    return os.path.join(os.path.expanduser("~"), ".config")


def get_settings_path(vendor_id, product_id):
    """Returns the path of a specific mouse settings file.

    :param int vendor_id: The device's vendor id (e.g. ``0x1038``).
    :param int product_id: The device's product id (e.g. ``0xbaad``).

    :rtype: str
    """
    return os.path.join(
        get_xdg_config_home(),
        "rivalcfg",
        "%04x_%04x.device.json"
        % (
            vendor_id,
            product_id,
        ),
    )


class MouseSettings(object):
    """Stores the settings of a mouse.

    :param int vendor_id: The device's vendor id (e.g. ``0x1038``).
    :param int product_id: The device's product id (e.g. ````0xbaad``).
    :param dict mouse_profile: The mouse profie (``devices/*``).
    :param str current_profile_name: The name of the active profile (optional,
                                     default: ``"default"``).
    """

    def __init__(
        self,
        vendor_id,
        product_id,
        mouse_profile,
        current_profile_name="default",
    ):
        self._mouse_profile = mouse_profile
        self._settings_path = get_settings_path(vendor_id, product_id)
        # TODO Check that the profile exists first!
        self._current_profile_name = current_profile_name
        self._settings = {}
        self._load()

    def list_settings_profiles(self):
        """List available profiles.

        .. WARNING::

            Not implemented yet!
        """
        raise NotImplementedError()  # TODO

    def set_active_profile(self, profile_name):
        """Change the active profile.

        .. WARNING::

            Not implemented yet!

        :param str profile_name: The name of the profile.
        """
        # TODO Raise error if the profile does not exist
        raise NotImplementedError()  # TODO

    def create_settings_profile(self, profile_name, from_profile_name=None):
        """Create a new setting profile.

        .. WARNING::

            Not implemented yet!

        :param str profile_name: The name of the new profile.
        :param str from_profile_name: Clone the given profile name (optional,
                                      default: ``None``).
        """
        # TODO Unique profile name (raise error if already exists)
        raise NotImplementedError()  # TODO

    def remove_settings_profile(self, profile_name):
        """Remove a profile.

        .. WARNING::

            Not implemented yet!

        :param str profile_name: The name of the profile to remove.
        """
        # TODO default profile cannot be removed
        # TODO Raise error if profile does not exist
        raise NotImplementedError()  # TODO

    def get_default_values(self):
        """Returns default settings of the device.

        :rtype: dict
        """
        return {
            k: v["default"]
            for k, v in self._mouse_profile["settings"].items()
            if "default" in self._mouse_profile["settings"][k]
        }

    def set(self, setting_name, value):
        """Set a value to a setting.

        :param str setting_name: The setting to set.
        :param value: The value to set.
        """
        if setting_name not in self._mouse_profile["settings"]:
            raise KeyError(
                "The %s device has no '%s' setting"
                % (
                    self._mouse_profile["name"],
                    setting_name,
                )
            )
        if (
            "value_type" in self._mouse_profile["settings"][setting_name]
            and self._mouse_profile["settings"][setting_name]["value_type"] == "none"
        ):
            return  # Skip settings without values
        self._settings[self._current_profile_name][setting_name] = value

    def get(self, setting_name):
        """Get the value of a setting.

        :param str setting_name: The setting to set.
        """
        if setting_name not in self._mouse_profile["settings"]:
            raise KeyError(
                "The %s device has no '%s' setting"
                % (
                    self._mouse_profile["name"],
                    setting_name,
                )
            )
        if (
            "value_type" in self._mouse_profile["settings"][setting_name]
            and self._mouse_profile["settings"][setting_name]["value_type"] == "none"
        ):
            return None
        return self._settings[self._current_profile_name][setting_name]

    def save(self):
        """Save settings in a file.

        .. NOTE::

            Settings are located in
            ``$XDG_CONFIG_HOME/rivalcfg/<vendor_id>_<product_id>.device.json``.
        """
        settings_dir = os.path.dirname(self._settings_path)
        if not os.path.isdir(settings_dir):
            os.makedirs(settings_dir)
        with open(self._settings_path, "w") as file_:
            json.dump(self._settings, file_, indent=2)

    def _load(self):
        """Load settings from a file.

        .. NOTE::

            Settings are located in
            ``$XDG_CONFIG_HOME/rivalcfg/<vendor_id>_<product_id>.device.json``.
        """
        if os.path.isfile(self._settings_path):
            with open(self._settings_path, "r") as file_:
                self._settings = json.load(file_)
        else:
            self._settings = {
                "default": self.get_default_values(),
            }


class FakeMouseSettings(MouseSettings):
    """An implementation of the :class:`MouseSettings` that does not make any I/O.

    :param int vendor_id: The device's vendor id (e.g. ``0x1038``).
    :param int product_id: The device's product id (e.g. ````0xbaad``).
    :param dict mouse_profile: The mouse profie (``devices/*``).
    :param str current_profile_name: The name of the active profile (optional,
                                     default: ``"default"``).
    """

    def save(self):
        pass

    def _load(self):
        self._settings = {
            "default": self.get_default_values(),
        }


def get_mouse_settings(
    vendor_id,
    product_id,
    mouse_profile,
    current_profile_name="default",
):
    """Returns a :class:`MouseSetting` instance.

    .. NOTE::

       A :class:`FakeMouseSettings` instance is returned when ``DEBUG_DRY``
       environment variable is set.

    :param int vendor_id: The device's vendor id (e.g. ``0x1038``).
    :param int product_id: The device's product id (e.g. ````0xbaad``).
    :param dict mouse_profile: The mouse profie (``devices/*``).
    :param str current_profile_name: The name of the active profile (optional,
                                     default: ``"default"``).

    :rtype MouseSettings, FakeMouseSettings
    """
    if "RIVALCFG_DRY" in os.environ:
        return FakeMouseSettings(
            vendor_id,
            product_id,
            mouse_profile,
            current_profile_name,
        )
    return MouseSettings(
        vendor_id,
        product_id,
        mouse_profile,
        current_profile_name,
    )



import pytest

from rivalcfg.__main__ import main


class TestMainCLI(object):
    def test_print_help(self):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            main(["--help"])
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0

        with pytest.raises(SystemExit) as pytest_wrapped_e:
            main(["-h"])
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0

    def test_print_version(self):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            main(["--version"])
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0

    def test_print_list(self):
        with pytest.raises(SystemExit) as pytest_wrapped_e:
            main(["--list"])
        assert pytest_wrapped_e.type == SystemExit
        assert pytest_wrapped_e.value.code == 0

    def test_no_save(self, monkeypatch):
        monkeypatch.setenv("RIVALCFG_DRY", "1")
        monkeypatch.setenv("RIVALCFG_PROFILE", "1038:1702")
        main(["--reset"])
        main(["-r"])
        # just expecting to have no error here... to boring to monkeypatch
        # the module to check if realy work...



"""
The "reactive_rgbcolor" type handles RGB color effects in reaction to a button
click.

It can be disabled with the following values:

* ``None``
* ``"off"``
* ``"disable"``

It supports hexadecimal colors:

* ``#FF0000``
* ``FF0000``
* ``#F00``
* ``F00``

RGB tuples or list (only from Python API for this one):

* ``(255, 0, 0)``
* ``[255, 0, 0]``

and named colors:

+------------+------------+-----------+-------------+
| ``white``  | ``red``    | ``lime``  | ``blue``    |
+------------+------------+-----------+-------------+
| ``silver`` | ``maroon`` | ``green`` | ``navy``    |
+------------+------------+-----------+-------------+
| ``gray``   | ``yellow`` | ``aqua``  | ``fuchsia`` |
+------------+------------+-----------+-------------+
| ``black``  | ``olive``  | ``teal``  | ``purple``  |
+------------+------------+-----------+-------------+


Device Profile
--------------

Example of a rgbcolor value type in a device profile:

::

    profile = {

        # ...

        "settings": {

            "reactive_color": {
                "label": "Reactive color",
                "description": "Set the color of the LED in reaction to a button click",
                "cli": ["-a", "--reactive-color"],
                "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
                "command": [0x26],
                "value_type": "reactive_rgbcolor",
                "default": "off",
            },

        },

        # ...

    }


CLI
---

Example of CLI option generated with this handler::

    -a REACTIVE_COLOR, --reactive-color REACTIVE_COLOR
                          Set the color of the LEDs in reaction to a button click
                          (e.g. off, disable, red, #ff0000, ff0000, #f00, f00,
                          default: off)

Example of CLI usage::

    rivalcfg --reactive-color=off
    rivalcfg --reactive-color=disable
    rivalcfg --reactive-color=red
    rivalcfg --reactive-color=ff0000
    rivalcfg --reactive-color=f00


Functions
---------
"""


import argparse

from ..color_helpers import is_color, parse_color_string


def process_value(setting_info, color):
    """Called by the :class:`rivalcfg.mouse.Mouse` class when processing a
    "reactive_rgbcolor" type setting.

    :param dict setting_info: The information dict of the setting from the
                              device profile.
    :param str,tuple,list,None color: The reactive color.
    :rtype: [int, int, int]
    """
    # Disable the reactive color
    if color is None or str(color).lower() in ["off", "disable"]:
        return [0x00, 0x00, 0x00, 0x00, 0x00]

    # Color tuple
    if type(color) in (tuple, list):
        if len(color) != 3:
            raise ValueError("Not a valid color %s" % str(color))
        for channel in color:
            if type(channel) != int or channel < 0 or channel > 255:
                raise ValueError("Not a valid color %s" % str(color))
        return [0x01, 0x00] + list(color)
    if is_color(color):
        return [0x01, 0x00] + list(parse_color_string(color))

    raise ValueError("Not a valid color %s" % str(color))


class CheckColorAction(argparse.Action):
    """Validate reactive colors from CLI"""

    def __call__(self, parser, namespace, value, option_string=None):
        if not is_color(value) and value.lower() not in ["off", "disable"]:
            raise argparse.ArgumentError(self, "invalid reactive color: '%s'" % value)
        setattr(namespace, self.dest.upper(), value)


def add_cli_option(cli_parser, setting_name, setting_info):
    """Add the given "reactive_rgbcolor" type setting to the given CLI
    arguments parser.

    :param ArgumentParser cli_parser: An :class:`ArgumentParser` instance.
    :param str setting_name: The name of the setting.
    :param dict setting_info: The information dict of the setting from the
                              device profile.
    """
    description = (
        "%s (e.g. off, disable, red, #ff0000, ff0000, #f00, f00, default: %s)"
        % (
            setting_info["description"],
            str(setting_info["default"]),
        )
    )
    cli_parser.add_argument(
        *setting_info["cli"],
        dest=setting_name,
        help=description,
        type=str,
        action=CheckColorAction,
        metavar=setting_name.upper()
    )



"""
The "multidpi_range" type alows to pick values in an input range and transforms
it into values from an output range. If the input values do not correspond to
one of the input range step, they are rounded to match the nearest step.

This type is used to support devices where we can configure from ``1`` to ``n``
DPI settings in a single command, like the Rival 3.

For example with an input range like ``[0, 1000, 100]`` and an output range
like ``[0, 10, 1]``, you can have the following pair:

* ``0`` -> ``0``
* ``100`` -> ``1``
* ``110`` -> ``1`` (``110`` rounded to ``100``)
* ``190`` -> ``2`` (``190`` rounded to ``200``)
* ``200`` -> ``2``
* ``300`` -> ``3``
* ...
* ``1000`` -> ``10``


Device Profile
--------------

Example of a multidpi_range value type in a device profile:

::

    profile = {

        # ...

        "settings": {

            "sensitivity1": {
                "label": "Sensitivity presets",
                "description": "Set sensitivity presets (DPI)",
                "cli": ["-s", "--sensitivity"],
                "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
                "command": [0x0B, 0x00],
                "value_type": "multidpi_range",
                "input_range": [200, 7200, 100],
                "output_range": [0x04, 0xA7, 2],
                "dpi_length_byte": 1,     # Little endian
                "first_preset": 1,
                "count_mode": "number",   # number, flag
                "max_preset_count": 5,
                "default": "800, 1600",
            },

        },

        # ...

    }


CLI
---

Example of CLI option generated with this handler::

   -s SENSITIVITY, --sensitivity SENSITIVITY
                        Set sensitivity preset (DPI) (up to 5 settings, from 200 dpi to 8500
                        dpi, default: '800, 1600')

Example of CLI usage::

    rivalcfg --sensitivity 1600
    rivalcfg --sensitivity "800,1600"
    rivalcfg --sensitivity "200, 400, 800, 1600, 8000"


Functions
---------
"""


import re
import argparse

from ..helpers import merge_bytes, uint_to_little_endian_bytearray
from .range import process_range


def process_value(setting_info, value, selected_preset=None):
    """Called by the :class:`rivalcfg.mouse.Mouse` class when processing a
    "range" type setting.

    :param dict setting_info: The information dict of the setting from the
                              device profile.
    :param value: The input value.
    :param int selected_preset: The DPI preset to select (0 is always the
                                first preset).
    :rtype: list[int]
    """
    dpis = []

    if type(value) in (int, float):
        dpis = [int(value)]
    elif type(value) in (list, tuple):
        dpis = [int(dpi) for dpi in value]
    else:
        dpis = [int(dpi) for dpi in value.replace(" ", "").split(",")]

    # Selected preset

    if selected_preset is None:
        selected_preset = setting_info["first_preset"]
    else:
        selected_preset += setting_info["first_preset"]

    # checks

    if len(dpis) == 0:
        raise ValueError("you must provide at least one preset")

    if len(dpis) > setting_info["max_preset_count"]:
        raise ValueError(
            "you provided %i preset but the device accepts a maximum of %i presets"
            % (len(dpis), setting_info["max_preset_count"])
        )

    if "first_preset" not in setting_info:
        raise ValueError(
            "Missing 'first_preset' parameter for 'multidpi_range' handler"
        )

    if (
        not setting_info["first_preset"]
        <= selected_preset
        < len(dpis) + setting_info["first_preset"]
    ):
        raise ValueError("the selected preset is out of range")

    if "dpi_length_byte" not in setting_info:
        raise ValueError(
            "Missing 'dpi_length_byte' parameter for 'multidpi_range' handler"
        )

    if "count_mode" not in setting_info:
        raise ValueError("Missing 'count_mode' parameter for 'multidpi_range' handler")

    if setting_info["count_mode"] not in ("number", "flag"):
        raise ValueError(
            "Invalid 'count_mode' parameter '%s'" % setting_info["count_mode"]
        )

    dpi_length = setting_info["dpi_length_byte"]
    count_mode = setting_info["count_mode"]

    # DPIs

    output_values = []

    for dpi in dpis:
        value = process_range(setting_info, dpi)
        value = uint_to_little_endian_bytearray(value, dpi_length)
        output_values = merge_bytes(output_values, value)

    # Count

    dpi_count = len(dpis)

    if count_mode == "flag":
        dpi_count = 0b11111111 >> (8 - dpi_count)

    #

    return merge_bytes(dpi_count, selected_preset, output_values)


def cli_multirange_validator(max_preset_count):
    class CheckMultiDpiRange(argparse.Action):
        """Validate value from CLI"""

        def __call__(self, parser, namespace, value, option_string=None):
            if not re.match(
                r"^ *[0-9]+( *, *[0-9]+){0,%i} *$" % (max_preset_count - 1),
                value,
            ):
                raise argparse.ArgumentError(self, "invalid DPI list: '%s'" % value)
            setattr(namespace, self.dest.upper(), value)

    return CheckMultiDpiRange


def add_cli_option(cli_parser, setting_name, setting_info):
    """Add the given "range" type setting to the given CLI arguments parser.

    :param ArgumentParser cli_parser: An :class:`ArgumentParser` instance.
    :param str setting_name: The name of the setting.
    :param dict setting_info: The information dict of the setting from the
                              device profile.
    """
    description = "%s (up to %i settings, from %i dpi to %i dpi, default: '%s')" % (
        setting_info["description"],
        setting_info["max_preset_count"],
        setting_info["input_range"][0],
        setting_info["input_range"][1],
        str(setting_info["default"]),
    )
    cli_parser.add_argument(
        *setting_info["cli"],
        help=description,
        dest=setting_name.upper(),
        metavar=setting_name.upper(),
        action=cli_multirange_validator(setting_info["max_preset_count"])
    )



from . import rival110


# This device looks like a Rival 300 but works like a Rival 110
profile = {
    "name": "SteelSeries Rival 300S",
    "models": [
        {
            "name": "SteelSeries Rival 300S",
            "vendor_id": 0x1038,
            "product_id": 0x1810,
            "endpoint": 0,
        }
    ],
    "settings": rival110.profile["settings"],
    "save_command": rival110.profile["save_command"],
}



from .. import usbhid


profile = {
    "name": "SteelSeries Rival 3",
    "models": [
        {
            "name": "SteelSeries Rival 3",
            "vendor_id": 0x1038,
            "product_id": 0x1824,
            "endpoint": 3,
        },
        {
            "name": "SteelSeries Rival 3 (firmware v0.37.0.0)",
            "vendor_id": 0x1038,
            "product_id": 0x184C,
            "endpoint": 3,
        },
    ],
    "settings": {
        "sensitivity": {
            "label": "Sensibility presets",
            "description": "Set sensitivity preset (DPI)",
            "cli": ["-s", "--sensitivity"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x0B, 0x00],
            "value_type": "multidpi_range",
            "input_range": [200, 8500, 100],
            "output_range": [4, 0xC5, 2.33],
            "dpi_length_byte": 1,
            "first_preset": 1,
            "count_mode": "number",
            "max_preset_count": 5,
            "default": "800, 1600",
        },
        "polling_rate": {
            "label": "Polling rate",
            "description": "Set polling rate (Hz)",
            "cli": ["-p", "--polling-rate"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x04, 0x00],
            "value_type": "choice",
            "choices": {
                125: 0x04,
                250: 0x03,
                500: 0x02,
                1000: 0x01,
            },
            "default": 1000,
        },
        # 0x05 0x00 <LED_ID> <R> <G> <B> <BRIGHTNESS>
        # LED_ID: 0x00 all
        #         0x01 zone 1 (top)
        #         0x02 zone 2 (middle)
        #         0x03 zone 3 (bottom)
        #         0x04 zone 4 / logo
        # BRIGHTNESS 0x00 - 0x64 (0 - 100%)
        "z1_color": {
            "label": "Strip top LED color",
            "description": "Set the color of the top LED of the strip",
            "cli": ["--strip-top-color", "--z1"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x05, 0x00, 0x01],
            "command_suffix": [0x64],
            "value_type": "rgbcolor",
            "default": "red",
        },
        "z2_color": {
            "label": "Strip middle LED color",
            "description": "Set the color of the middle LED of the strip",
            "cli": ["--strip-middle-color", "--z2"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x05, 0x00, 0x02],
            "command_suffix": [0x64],
            "value_type": "rgbcolor",
            "default": "lime",
        },
        "z3_color": {
            "label": "Strip bottom LED color",
            "description": "Set the color of the bottom LED of the strip",
            "cli": ["--strip-bottom-color", "--z3"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x05, 0x00, 0x03],
            "command_suffix": [0x64],
            "value_type": "rgbcolor",
            "default": "blue",
        },
        "logo_color": {
            "label": "Logo LED color",
            "description": "Set the color of the logo LED",
            "cli": ["-c", "--logo-color", "--z4"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x05, 0x00, 0x04],
            "command_suffix": [0x64],
            "value_type": "rgbcolor",
            "default": "purple",
        },
        "light_effect": {
            "label": "Light effect",
            "description": "Set the light effect",
            "cli": ["-e", "--light-effect"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x06, 0x00],
            "value_type": "choice",
            "choices": {
                "rainbow-shift": 0x00,
                "breath-fast": 0x01,
                "breath": 0x02,
                "breath-slow": 0x03,
                "steady": 0x04,
                "rainbow-breath": 0x05,
                "disco": 0x06,
            },
            "default": "steady",
        },
        "buttons_mapping": {
            "label": "Buttons mapping",
            "description": "Set the mapping of the buttons",
            "cli": ["-b", "--buttons"],
            "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
            "command": [0x07, 0x00],
            "value_type": "buttons",
            # fmt: off
            "buttons": {
                "Button1":    {"id": 0x01, "offset": 0x00, "default": "button1"},
                "Button2":    {"id": 0x02, "offset": 0x02, "default": "button2"},
                "Button3":    {"id": 0x03, "offset": 0x04, "default": "button3"},
                "Button4":    {"id": 0x04, "offset": 0x06, "default": "button4"},
                "Button5":    {"id": 0x05, "offset": 0x08, "default": "button5"},
                "Button6":    {"id": 0x06, "offset": 0x0A, "default": "dpi"},
                "ScrollUp":   {"id": 0x31, "offset": 0x0C, "default": "scrollup"},
                "ScrollDown": {"id": 0x32, "offset": 0x0E, "default": "scrolldown"},
            },
            "button_disable":     0x00,
            "button_dpi_switch":  0x30,
            "button_scroll_up":   0x31,
            "button_scroll_down": 0x32,
            "button_keyboard":    0x33,
            "button_multimedia":  0x34,
            # fmt: on
            "button_field_length": 2,
            "default": "buttons(button1=button1; button2=button2; button3=button3; button4=button4; button5=button5; button6=dpi; scrollup=scrollup; scrolldown=scrolldown; layout=qwerty)",
        },
    },
    "save_command": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "command": [0x09, 0x00],
    },
    "firmware_version": {
        "report_type": usbhid.HID_REPORT_TYPE_OUTPUT,
        "command": [0x10, 0x00],
        "response_length": 2,
    },
}
